#ifdef GLBACKEND_SDL


#include <SDL2/SDL.h>
#include <SDL2/SDL_video.h>


constexpr auto width = 640;
constexpr auto height = 480;


int main()
{
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
        return 1;
    auto * window = SDL_CreateWindow(
        "baseline_sdl",
        SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, // NOLINT
        width, height,
        SDL_WINDOW_OPENGL
    );
    if (!window)
        return 1;
    SDL_DestroyWindow(window);
    SDL_Quit();
}


#else
int main() {}
#endif
