#ifndef GLBACKEND_DEFAULT_HPP_
#define GLBACKEND_DEFAULT_HPP_


#if defined(GLBACKEND_GLFW)
    #include "glbackend_glfw.hpp" // IWYU pragma: export
    using GLBackendDefault = GLBackendGLFW;
#elif defined(GLBACKEND_SDL)
    #include "glbackend_sdl.hpp" // IWYU pragma: export
    using GLBackendDefault = GLBackendSDL;
#elif defined(GLBACKEND_WXWIDGETS)
    #include "glbackend_wxwidgets.hpp" // IWYU pragma: export
    using GLBackendDefault = GLBackendWxWidgets;
#else
    #error "No GLBackendDefault configured"
#endif


#endif
