#ifdef GLBACKEND_GLFW


#include <GLFW/glfw3.h>


constexpr auto width = 640;
constexpr auto height = 480;


int main()
{
    if (!glfwInit())
        return 1;
    auto * window = glfwCreateWindow(
        width, height,
        "baseline_glfw",
        nullptr, nullptr
    );
    if (!window)
        return 1;
    glfwDestroyWindow(window);
    glfwTerminate();
}


#else
int main() {}
#endif
